#ifndef _CHRONOPTICS_TOF_GIGE_INTERFACE_HPP_
#define _CHRONOPTICS_TOF_GIGE_INTERFACE_HPP_

#include <chronoptics/tof/gige_interface.h>

#include <chronoptics/tof/camera_config.hpp>
#include <chronoptics/tof/calibration.hpp>
#include <chronoptics/tof/data.hpp>

namespace chronoptics {
namespace tof {

/** The possible depth streams that can be streamed from the camera
 */
enum class DepthStreamType {
  RAW = 0, /** Stream raw data */
  RADIAL_INTENSITY = 1, /** Stream both radial and intensity */
  RADIAL = 2, /** Stream radial */
  INTENSITY = 3, /** Stream intensity */
};

/** A gige discovery message containing camera information
*/
class DiscoveryMessage : public detail::Base<tof_discovery_message, tof_discovery_message_delete> {
 public:
  /** Construct from pointer */
  DiscoveryMessage(tof_discovery_message_t ptr = nullptr) {
    this->ptr_ = ptr;
  }

  /** Get the camera's ip address
   * @return Ip address
   */
  const char* ip() const {
    return tof_discovery_message_ip(this->ptr_, TOF_ERROR_HANDLER{});
  }

  /** Get the camera's serial
   * @return Serial
   */
  const char* serial() const {
    return tof_discovery_message_serial(this->ptr_, TOF_ERROR_HANDLER{});
  }

};

/** A persistent ip data structure
*/
class PersistentIp : public detail::Base<tof_persistent_ip, tof_persistent_ip_delete> {
 public:
  /** Construct from pointer */
  PersistentIp(tof_persistent_ip_t ptr = nullptr) {
    this->ptr_ = ptr;
  }

  /** Create persistent ip
   * @param ip IP address
   * @param netmask Netmask
   * @param gateway Gateway
   * @param enabled Enable or disable persistent ip
   */
  PersistentIp(StringView ip, StringView netmask, StringView gateway, bool enabled) {
    this->ptr_ = tof_persistent_ip_new(ip, netmask, gateway, enabled, TOF_ERROR_HANDLER{});
  }

  /** Get the IP address
   * @return the IP address
   */
  const char* get_ip() const {
    return tof_persistent_ip_get_ip(this->ptr_, TOF_ERROR_HANDLER{});
  }

  /** Set the IP address
   * @param ip the IP address
   */
  void set_ip(StringView ip) {
    return tof_persistent_ip_set_ip(this->ptr_, ip, TOF_ERROR_HANDLER{});
  }

  /** Get the netmask
   * @return The netmask
   */
  const char* get_netmask() const {
    return tof_persistent_ip_get_netmask(this->ptr_, TOF_ERROR_HANDLER{});
  }

  /** Set the netmask
   * @param netmask The netmask
   */
  void set_netmask(StringView netmask) {
    return tof_persistent_ip_set_netmask(this->ptr_, netmask, TOF_ERROR_HANDLER{});
  }

  /** Get the gateway
   * @return The gateway
   */
  const char* get_gateway() const {
    return tof_persistent_ip_get_gateway(this->ptr_, TOF_ERROR_HANDLER{});
  }

  /** Set the gateway
   * @param gateway The gateway
   */
  void set_gateway(StringView gateway) {
    return tof_persistent_ip_set_gateway(this->ptr_, gateway, TOF_ERROR_HANDLER{});
  }

  /** Get if persistent ip is enabled
   * @return Is enabled
   */
  bool get_enabled() const {
    return tof_persistent_ip_get_enabled(this->ptr_, TOF_ERROR_HANDLER{});
  }

  /** Set if persistent ip is enabled
   * @param enabled Is enabled
   */
  void set_enabled(bool enabled) {
    return tof_persistent_ip_set_enabled(this->ptr_, enabled, TOF_ERROR_HANDLER{});
  }

};

/** Connect to a camera directly using the gige interface
*/
class GigeInterface : public detail::Base<tof_gige_interface, tof_gige_interface_delete> {
 public:
  /** Construct from pointer */
  GigeInterface(tof_gige_interface_t ptr) {
    this->ptr_ = ptr;
  }

  /** Create an instance of the gige interface
   * @param port The port to use for GVCP communication, if 0 is specified it
   * will use an OS assigned port
   */
  GigeInterface(uint16_t port = 0) {
    this->ptr_ = tof_gige_interface_new(port, TOF_ERROR_HANDLER{});
  }

  /** Detect all the cameras on the network
   * @return A list of discovery messages
   */
  std::vector<DiscoveryMessage> discover() {
    size_t size = tof_gige_interface_discover(this->ptr_, nullptr, 0, TOF_ERROR_HANDLER{});
    std::vector<DiscoveryMessage> vec;
    vec.reserve(size);
    for (size_t i = 0; i < size; i++)
      vec.emplace_back(static_cast<tof_discovery_message_t>(nullptr));
    auto data = reinterpret_cast<tof_discovery_message_t*>(vec.data());
    size = tof_gige_interface_discover(this->ptr_, data, vec.size(), TOF_ERROR_HANDLER{});
    return vec;
  }

  /** Returns a message of the first camera that responded
   * @return A discovery message
   */
  DiscoveryMessage discover_one() {
    DiscoveryMessage new_discovery_message(static_cast<tof_discovery_message_t>(nullptr));
    auto ptr = reinterpret_cast<tof_discovery_message_t*>(&new_discovery_message);
    *ptr = tof_gige_interface_discover_one(this->ptr_, TOF_ERROR_HANDLER{});
    return new_discovery_message;
  }

  /** Detect a specific camera on the network, will throw an error if it can't
   * find the camera
   * @param serial Serial number of camera to find
   * @return A discovery message
   */
  DiscoveryMessage find(StringView serial) {
    DiscoveryMessage new_discovery_message(static_cast<tof_discovery_message_t>(nullptr));
    auto ptr = reinterpret_cast<tof_discovery_message_t*>(&new_discovery_message);
    *ptr = tof_gige_interface_find(this->ptr_, serial, TOF_ERROR_HANDLER{});
    return new_discovery_message;
  }

  /** Connect to the camera and gain exclusive access
   * @param message Discovery message of the camera to connect to
   */
  void connect(const DiscoveryMessage &message) {
    return tof_gige_interface_connect(this->ptr_, *reinterpret_cast<const tof_discovery_message_t*>(&message), TOF_ERROR_HANDLER{});
  }

  /** Disconnect from the camera
   */
  void disconnect() {
    return tof_gige_interface_disconnect(this->ptr_, TOF_ERROR_HANDLER{});
  }

  /** Check whether a connection is still maintained with the camera
   * @return Is connected
   */
  bool is_connected() const {
    return tof_gige_interface_is_connected(this->ptr_, TOF_ERROR_HANDLER{});
  }

  /** Get stream packet size
   * @return The packet size
   */
  uint16_t get_stream_packet_size() const {
    return tof_gige_interface_get_stream_packet_size(this->ptr_, TOF_ERROR_HANDLER{});
  }

  /** Set stream packet size
   * @param packet_size The packet size
   */
  void set_stream_packet_size(uint16_t packet_size) {
    return tof_gige_interface_set_stream_packet_size(this->ptr_, packet_size, TOF_ERROR_HANDLER{});
  }

  /** Test whether using the set packet size works
   * @param port Port to send test packets to
   * @return Packet size works
   */
  bool test_stream_packet_size(uint16_t port = 0) {
    return tof_gige_interface_test_stream_packet_size(this->ptr_, port, TOF_ERROR_HANDLER{});
  }

  /** Get delay between network packets
   * @return The delay, depending on the camera software it is in nanoseconds or
   * in cpu instruction
   */
  uint32_t get_delay() const {
    return tof_gige_interface_get_delay(this->ptr_, TOF_ERROR_HANDLER{});
  }

  /** Set delay between network packets
   * @param delay The delay, depending on the camera software it is in
   * nanoseconds or in cpu instruction
   */
  void set_delay(uint32_t delay) {
    return tof_gige_interface_set_delay(this->ptr_, delay, TOF_ERROR_HANDLER{});
  }

  /** Download camera configuration from the camera
   * @param port Port to download configuration on
   * @return Camera configuration
   */
  CameraConfig download_configuration(uint16_t port = 0) {
    CameraConfig new_camera_config(static_cast<tof_camera_config_t>(nullptr));
    auto ptr = reinterpret_cast<tof_camera_config_t*>(&new_camera_config);
    *ptr = tof_gige_interface_download_configuration(this->ptr_, port, TOF_ERROR_HANDLER{});
    return new_camera_config;
  }

  /** Download calibration from the camera
   * @param port Port to download configuration on
   * @return Calibration
   */
  Calibration download_calibration(uint16_t port = 0) {
    Calibration new_calibration(static_cast<tof_calibration_t>(nullptr));
    auto ptr = reinterpret_cast<tof_calibration_t*>(&new_calibration);
    *ptr = tof_gige_interface_download_calibration(this->ptr_, port, TOF_ERROR_HANDLER{});
    return new_calibration;
  }

  /** Upload camera configuration to the camera
   * @param config Camera configuration
   * @param port Port to download configuration on
   */
  void upload_configuration(const CameraConfig &config, uint16_t port = 0) {
    return tof_gige_interface_upload_configuration(this->ptr_, *reinterpret_cast<const tof_camera_config_t*>(&config), port, TOF_ERROR_HANDLER{});
  }

  /** Check whether this camera is capable of streaming the specified type
   * @param stream Depth stream type
   * @return Is capable
   */
  bool depth_stream_capable(DepthStreamType stream) const {
    return tof_gige_interface_depth_stream_capable(this->ptr_, static_cast<tof_depth_stream_type>(stream), TOF_ERROR_HANDLER{});
  }

  /** Start the depth stream
   * @param stream Depth stream type
   * @param port Port to receive depth data on
   */
  void start_depth_stream(DepthStreamType stream, uint16_t port = 0) {
    return tof_gige_interface_start_depth_stream(this->ptr_, static_cast<tof_depth_stream_type>(stream), port, TOF_ERROR_HANDLER{});
  }

  /** Stop the depth stream
   */
  void stop_depth_stream() {
    return tof_gige_interface_stop_depth_stream(this->ptr_, TOF_ERROR_HANDLER{});
  }

  /** Returns true if a depth frame is available
   * @return Depth frame available
   */
  bool has_depth_data() const {
    return tof_gige_interface_has_depth_data(this->ptr_, TOF_ERROR_HANDLER{});
  }

  /** Get a depth frame
   * @return Depth frame
   */
  Data get_depth_data() {
    Data new_data(static_cast<tof_data_t>(nullptr));
    auto ptr = reinterpret_cast<tof_data_t*>(&new_data);
    *ptr = tof_gige_interface_get_depth_data(this->ptr_, TOF_ERROR_HANDLER{});
    return new_data;
  }

  /** Check whether camera supports an image stream
   * @return Camera supports an image stream
   */
  bool image_stream_capable() const {
    return tof_gige_interface_image_stream_capable(this->ptr_, TOF_ERROR_HANDLER{});
  }

  /** Start image stream
   * @param port Port to receive image data on
   */
  void start_image_stream(uint16_t port = 0) {
    return tof_gige_interface_start_image_stream(this->ptr_, port, TOF_ERROR_HANDLER{});
  }

  /** Stop image stream
   */
  void stop_image_stream() {
    return tof_gige_interface_stop_image_stream(this->ptr_, TOF_ERROR_HANDLER{});
  }

  /** Check whether image frame is available
   * @return Image frame available
   */
  bool has_image_data() const {
    return tof_gige_interface_has_image_data(this->ptr_, TOF_ERROR_HANDLER{});
  }

  /** Get image frame
   * @return Image data
   */
  Data get_image_data() {
    Data new_data(static_cast<tof_data_t>(nullptr));
    auto ptr = reinterpret_cast<tof_data_t*>(&new_data);
    *ptr = tof_gige_interface_get_image_data(this->ptr_, TOF_ERROR_HANDLER{});
    return new_data;
  }

  /** Get the image frame into the supplied pointer
   * @param pointer The pointer to the data to write into
   * @param capacity The amount of data the pointer can hold
   * @param callback Callback that will be called when pointer is no longer in
   * use
   * @return Image data
   */
  Data get_image_data_into_pointer(uint8_t* pointer, size_t capacity, user_pointer_destructed_fn &callback) {
    Data new_data(static_cast<tof_data_t>(nullptr));
    auto ptr = reinterpret_cast<tof_data_t*>(&new_data);
    *ptr = tof_gige_interface_get_image_data_into_pointer(this->ptr_, pointer, capacity, user_pointer_destructed_cb_handler, &callback, TOF_ERROR_HANDLER{});
    return new_data;
  }

  /** Get the image frame into the supplied pointer
   * @param pointer The pointer to the data to write into
   * @param capacity The amount of data the pointer can hold
   * @param callback Callback that will be called when pointer is no longer in
   * use
   * @param callback_user_data User data that will be passed back when the function pointer is called
   * @return Image data
   */
  Data get_image_data_into_pointer(uint8_t* pointer, size_t capacity, tof_user_pointer_destructed_t callback, void* callback_user_data) {
    Data new_data(static_cast<tof_data_t>(nullptr));
    auto ptr = reinterpret_cast<tof_data_t*>(&new_data);
    *ptr = tof_gige_interface_get_image_data_into_pointer(this->ptr_, pointer, capacity, callback, callback_user_data, TOF_ERROR_HANDLER{});
    return new_data;
  }

  /** Get whether the camera is persistent ip capable
   * @return Persistent IP capable
   */
  bool persistent_ip_capable() const {
    return tof_gige_interface_persistent_ip_capable(this->ptr_, TOF_ERROR_HANDLER{});
  }

  /** Get the current persistent ip settings
   * @return Persistent IP
   */
  PersistentIp get_persistent_ip() const {
    PersistentIp new_persistent_ip(static_cast<tof_persistent_ip_t>(nullptr));
    auto ptr = reinterpret_cast<tof_persistent_ip_t*>(&new_persistent_ip);
    *ptr = tof_gige_interface_get_persistent_ip(this->ptr_, TOF_ERROR_HANDLER{});
    return new_persistent_ip;
  }

  /** Set the persistent ip settings. The persistent ip settings will be active
   * on next reboot.
   * @param persistent_ip The persistent ip
   */
  void set_persistent_ip(const PersistentIp &persistent_ip) {
    return tof_gige_interface_set_persistent_ip(this->ptr_, *reinterpret_cast<const tof_persistent_ip_t*>(&persistent_ip), TOF_ERROR_HANDLER{});
  }

  /** Get the capacity for amount of user pointers that can be registered
   * @return User pointer capacity
   */
  size_t get_user_pointer_capacity() const {
    return tof_gige_interface_get_user_pointer_capacity(this->ptr_, TOF_ERROR_HANDLER{});
  }

  /** Set the capacity for amount of user pointers that can be registered
   * @param capacity User pointer capacity
   */
  void set_user_pointer_capacity(size_t capacity) {
    return tof_gige_interface_set_user_pointer_capacity(this->ptr_, capacity, TOF_ERROR_HANDLER{});
  }

  /** Add a pointer that will be filled with depth data. This method allows you
   * to fill data straight into your own data structure without any additional
   * copying.
   * @param pointer The pointer to the data to write into
   * @param capacity The amount of data the pointer can hold
   * @param callback Callback that will be called when pointer is no longer in
   * use
   */
  void add_depth_user_pointer(uint8_t* pointer, size_t capacity, user_pointer_destructed_fn &callback) {
    return tof_gige_interface_add_depth_user_pointer(this->ptr_, pointer, capacity, user_pointer_destructed_cb_handler, &callback, TOF_ERROR_HANDLER{});
  }

  /** Add a pointer that will be filled with depth data. This method allows you
   * to fill data straight into your own data structure without any additional
   * copying.
   * @param pointer The pointer to the data to write into
   * @param capacity The amount of data the pointer can hold
   * @param callback Callback that will be called when pointer is no longer in
   * use
   * @param callback_user_data User data that will be passed back when the function pointer is called
   */
  void add_depth_user_pointer(uint8_t* pointer, size_t capacity, tof_user_pointer_destructed_t callback, void* callback_user_data) {
    return tof_gige_interface_add_depth_user_pointer(this->ptr_, pointer, capacity, callback, callback_user_data, TOF_ERROR_HANDLER{});
  }

  /** Get the tof library version running on the camera
   * @return Version
   */
  const char* version() {
    return tof_gige_interface_version(this->ptr_, TOF_ERROR_HANDLER{});
  }

};

} // tof
} // chronoptics

#endif
